package drastic.mCmdIF;

import java.net.*;
import java.util.*;
import java.io.*;

/**
 * <p>Title: SocketManager</p>
 * 
 * <p>Description: Track sockets for reuse by multiple callers.  Uses the
 * SocketData class to manage/track each individual socket.  Do not call directly,
 * use MediaCmdIF which uses VVWNetCmdSender to communicate with this
 * SocketManager.</p>
 *
 * <p>Copyright: Copyright  2004-2007</p>
 *
 * <p>Company: Drastic Technologies Ltd.<br>
 * 523 The Queensway, Suite 102<br>
 * Toronto, ON, M8Y 1J7<br>
 * CANADA<br>
 * 416 255 5636<br>
 * fax 255 8780<br>
 * http://www.drastictech.com<br>
 * engineering@drastictech.com</p>
 *
 * @see SocketData
 * @see VVWNETCMDSENDER
 * @author Dean Mitrovich
 * @version 1.0
 */
public class SocketManager
{
    /**
     * <p>Title: SocketData</p>
     * <p>Description: Single socket within a socket manager.   Used by the
     * socket manager to manage each unique socket it creates.  This way there
     * is only one actual connection to each VVW no matter how many logical
     * connections, or per internal channel connections are made.</p>
     * <p>Copyright: Copyright  2004-2007</p>
     * <p>Company: Drastic Technologies Ltd</p>
     * @see SocketManager
     * @author Dean Mitrovich
     * @version 1.0
     */
    class SocketData
    {
        /**
         * Number of callers using this socket
         */
        long dwRefCount;
        /**
         * The actuall java socket we are using for communications
         */
        Socket socketUsed;
        /**
         * The remote address we are connected to
         */
        InetAddress inetIPAddress;
        /**
         * The port number we connected on
         */
        long dwPort;
        /**
         * Are we on our first connection
         */
        boolean fFirstConnect = true;

        /**
         * Only for SocketLocket, do not use outside this function.  One per socket.
         */
        private boolean m_fInUse_SocketLocket = false;

        /**
         * SocketLocket - Lock a socket for exclusive use
         *
         * @param fCheck boolean Do we want to acquire the lock, or just check if it is locked
         * @param fLock boolean Are we locking, or unlocking the socket
         * @return boolean true if locked
         */
        private synchronized boolean SocketLocket(boolean fCheck, boolean fLock)
        {
            // Just a check
            if(fCheck)
            {
                return m_fInUse_SocketLocket;
            }
            // Unlocking
            if(fLock == false)
            {
                m_fInUse_SocketLocket = false;
                return m_fInUse_SocketLocket;
            }
            // Locking
            if(m_fInUse_SocketLocket == true)
            {
                // Already locked
                return false;
            }
            // Lock it
            m_fInUse_SocketLocket = true;
            return m_fInUse_SocketLocket;
        }
        /**
         * fInUse - Is the socket currently locked and in use
         *
         * @return boolean true if it is in use
         */
        boolean fInUse()
        {
            return SocketLocket(true, false);
        }
        /**
         * LockSocket - Lock the socket for exclusive use
         *
         * @return boolean true if it is in use
         */
        boolean LockSocket()
        {
            return SocketLocket(false, true);
        }
        /**
         * UnLockSocket - Unlock the socket, allow others access
         *
         * @return boolean true if it is in use
         */
        boolean UnLockSocket()
        {
            return SocketLocket(false, false);
        }
    }

    /**
     * The list of active sockets - 
     * NOTE: To maintain backward compatibility, do NOT parameterize this
     */
    LinkedList SocketList = new LinkedList();
    /**
     * Length of time in milliseconds to wait for a socket to connect.
     */
    private int m_timeOut =  500; // 2000;

    /**
     * SocketManager
     */
    public SocketManager()
    {
    }

    /**
     * Find a socket that has already been opened.  If there is a socket
     * connected to inetAddress on that dwPort then return it.  If not,
     * return null.
     *
     * @param inetAddress InetAddress the target machine address or name
     * @param dwPort int the port to connect on
     * @return SocketData either null or the socket if it exists
     */
    SocketData Find(InetAddress inetAddress, int dwPort)
    {
        SocketData tempSocketData;
        // NOTE: To maintain backward compatibility, do NOT parameterize this
        Iterator iter = SocketList.iterator();

        for(int i = 1; i <= SocketList.size(); i++)
        {
            tempSocketData = (SocketData) iter.next();

            if(tempSocketData.inetIPAddress.getHostAddress().compareTo(inetAddress.getHostAddress()) == 0 &&
                tempSocketData.dwPort == dwPort)
            {
                // Found the socket
                return tempSocketData;
            }
        }
        // This connection does not exist
        return null;
    }

    /**
     * Open a socket.  Either return an existing socket to that address and
     * port, or create a new one and return that.  New ones are added to the
     * list.
     *
     * @param inetAddress InetAddress the ip address or name to attach to
     * @param dwPort int the port to attach on
     * @return boolean true if socket is open, else open failed
     */
    boolean Open(InetAddress inetAddress, int dwPort)
    {
        InetSocketAddress inetSocketAddress;
        Socket newSocket;
        SocketData newSocketData;
        SocketData tempSocketData;
        newSocketData = new SocketData();
        // NOTE: To maintain backward compatibility, do NOT parameterize this
        Iterator iter = SocketList.iterator();

        if (SocketList.size() < 1)
        {
            // First element
            try
            {
                newSocket = new Socket();
                inetSocketAddress = new InetSocketAddress(inetAddress, dwPort);
                newSocket.connect(inetSocketAddress, m_timeOut);
            }
            catch (Exception e)
            {
                // Print any errors produced by above try statment
                System.out.println("SocketManager.Open() Failed: Exception: " + e.getMessage());
                return false;
            }

            if (newSocket == null)
            {
                return false;
            }

            newSocketData.dwRefCount = 1;

            newSocketData.socketUsed = newSocket;
            newSocketData.inetIPAddress = inetAddress;
            newSocketData.dwPort = dwPort;

            try
            {
                // enable TCP_NODELAY
                newSocketData.socketUsed.setTcpNoDelay(true);
                // Set a read timeout
                newSocketData.socketUsed.setSoTimeout(m_timeOut);
            }
            catch (Exception e)
            {
                // debug here
            }
            // NOTE: To maintain backward compatibility, do NOT parameterize this
            SocketList.add(newSocketData);
            System.out.println("New Socket Opened ");
        }
        else
        {
            // Subsequent elements
            for (int i = 1; i <= SocketList.size(); i++)
            {
                    tempSocketData = (SocketData) iter.next();

                    if (tempSocketData.inetIPAddress.getHostAddress().compareTo(inetAddress.getHostAddress()) == 0 &&
                        tempSocketData.dwPort == dwPort)
                    {
                        // Found existing socket for same address
                        tempSocketData.dwRefCount += 1;
                        tempSocketData.fFirstConnect = false;
                        // NOTE: To maintain backward compatibility, do NOT parameterize this
                        SocketList.set(i - 1, tempSocketData);
                        return true;
                    }
            }

            // No existing socket found for this connection, create a new one
            try
            {
                newSocket = new Socket(inetAddress, dwPort);
            }
            catch (Exception e)
            {
                // Print any errors produced by above try statment
                System.out.println("Exception: " + e.getMessage());
                return false;
            }

            if (newSocket == null)
            {
                return false;
            }

            newSocketData.dwRefCount = 1;

            newSocketData.socketUsed = newSocket;
            newSocketData.inetIPAddress = inetAddress;
            newSocketData.dwPort = dwPort;

            // NOTE: To maintain backward compatibility, do NOT parameterize this
            SocketList.add(newSocketData);
            System.out.println("Pseudo Socket Opened ");
       }
        return true;
    }

    /**
     * Returns whether we have a connection to this socket/port or not.
     *
     * @param inetAddress InetAddress the ip address or name to attach to
     * @param dwPort int the port to attach on
     * @return boolean true if we have an open socket for this name/port
     */
    boolean IsSocketConnected(InetAddress inetAddress, int dwPort)
    {
        SocketData tempSocketData = null;
        // NOTE: To maintain backward compatibility, do NOT parameterize this
        Iterator iter = SocketList.iterator();

        for (int i = 1; i <= SocketList.size(); i++)
        {
                tempSocketData = (SocketData) iter.next();

                if (tempSocketData.inetIPAddress.getHostAddress().compareTo(inetAddress.getHostAddress()) == 0 &&
                    tempSocketData.dwPort == dwPort)
                {
                    // Found this socket
                    if (tempSocketData.socketUsed.isConnected())
                    {
                        return true;
                    }
                }
        }
        return false;
    }

    /**
     * Changes the tpcip (SoTimeOut) timeout value for a socket.
     *
     * @param inetAddress InetAddress the ip address or name to attach to
     * @param dwPort int the port to attach on
     * @param nTimeOut int new time out value for the socket
     */
    void setTimeout(InetAddress inetAddress, int dwPort, int nTimeOut)
    {
        SocketData tempSocketData = null;
        // NOTE: To maintain backward compatibility, do NOT parameterize this
        Iterator iter = SocketList.iterator();

        for(int i = 1; i <= SocketList.size(); i++)
        {
            tempSocketData = (SocketData) iter.next();

            if(tempSocketData.inetIPAddress.getHostAddress().compareTo(inetAddress.getHostAddress()) == 0 &&
                tempSocketData.dwPort == dwPort)
            {
                try
                {
                    // Found the socket
                    tempSocketData.socketUsed.setSoTimeout(nTimeOut);
                } catch(Exception e)
                {
                }
            }
        }
    }

    /**
     * Close the socket.  If fAllInstances is false, will close the reference,
     * or if the reference is one it will close the actual socket.  If
     * fAllInstances is true, it will always close the actual socket.
     *
     * @param inetAddress InetAddress the ip address or name to close
     * @param dwPort int the port to close
     * @param fAllInstances boolean if true, close the actual socket.  if false,
     * reduce the reference count until it is 1, then close the actual socket
     * @return boolean true if socket was closed or de-referenced
     */
    boolean Close(InetAddress inetAddress, int dwPort, boolean fAllInstances)
    {
        SocketData tempSocketData = null;
        // NOTE: To maintain backward compatibility, do NOT parameterize this
        Iterator iter = SocketList.iterator();

        for (int i = 1; i <= SocketList.size(); i++)
        {
           tempSocketData = (SocketData) iter.next();

           if(tempSocketData.inetIPAddress.getHostAddress().compareTo(inetAddress.getHostAddress()) == 0 && tempSocketData.dwPort == dwPort)
           {
               // Found this socket
               if (tempSocketData.dwRefCount < 2 || fAllInstances)
               {
                   // Not being used elsewhere or flagged to close regardless
                   try
                   {
                      // tempSocketData.socketUsed.getInputStream().close();
                      // tempSocketData.socketUsed.getOutputStream().close();

                      tempSocketData.socketUsed.close();

                       if (tempSocketData.socketUsed.isClosed())
                       {
                           System.out.println("\n\rSocket was successfully closed");
                       }
                       else
                       {
                           System.out.println("\n\rSocket was NOT successfully closed");
                       }
                       tempSocketData.socketUsed = null;

                       // Remove last visited element
                       iter.remove();

                       return true;
                   }
                   catch (Exception e)
                   {
                       System.out.println("\n\rSocketManager.Close(): Exception: " + e.getMessage());
                   }
               }
               else
               {
                   // It's also being used elsewhere, so just decrement the RefCount
                   tempSocketData.dwRefCount--;
                   // NOTE: To maintain backward compatibility, do NOT parameterize this
                   SocketList.set(i-1, tempSocketData);

                   return true;
               } // end else
           }
        }
        return false; // true;
    }

    /**
     * Get the input stream.  Create and return an input stream for a socket.
     *
     * @param inetAddress InetAddress the ip address or name to attach to
     * @param dwPort int the port to attach on
     * @return InputStream input stream on the socket
     */
    InputStream GetInputStream(InetAddress inetAddress, int dwPort)
    {
        SocketData tempSocketData = null;
        InputStream strm = null;
        tempSocketData = Find(inetAddress, dwPort);

        if(tempSocketData != null)
        {
            try
            {
                strm = tempSocketData.socketUsed.getInputStream();
            } catch (IOException e)
            {
                // Print any errors produced by above try statment
                System.out.println("\n\rSocketManager.GetInputStream():IO Exception: " + e.getMessage());
                return null;
            }
        }
        return strm;
    }

    /**
     * Get the output stream.  Create and return an output stream for a socket.
     *
     * @param inetAddress InetAddress the ip address or name to attach to
     * @param dwPort int the port to attach on
     * @return OutputStream output stream on the socket
     */
    OutputStream GetOutputStream(InetAddress inetAddress, int dwPort)
    {
        SocketData tempSocketData = null;
        OutputStream strm = null;
        tempSocketData = Find(inetAddress, dwPort);

        if(tempSocketData != null)
        {
            try
            {
                strm = tempSocketData.socketUsed.getOutputStream();
            } catch (IOException e)
            {
                //Print any errors produced by above try statment
                System.out.println("\n\rIO Exception: " + e.getMessage());
            }
        }
        return strm;
    }
}
