package drastic.mCmdIF;

import java.net.*;
import java.util.Date;
import java.io.*;

/**
 * <p>Title: VVWNetCmdSender (low level - use MediaCmdIF instead)</p>
 *
 * <p>Description: Send VVW MediaCmd s through a network TCP-IP socket interface
 * to a remote VVW/Titan or network enabled QuickClip.  Handles the send and
 * receive through a direct socket, or through the socket manager.
 *<br><br>
 * MediaCmdIF - Main Interface API - uses MediaCmd and VVWNetCmdSender<br>
 * MediaCmd - Structure sent and returned from the VVW<br>
 * VVWNETCMDSENDER - used to send and receive the MediaCmds<br>
 * </p>
 *
 * <p>Company: Drastic Technologies Ltd.<br>
 * 523 The Queensway, Suite 102<br>
 * Toronto, ON, M8Y 1J7<br>
 * CANADA<br>
 * 416 255 5636<br>
 * fax 255 8780<br>
 * http://www.drastictech.com<br>
 * engineering@drastictech.com</p>
 *
 * @see MediaCmd
 * @see MediaCmdIF
 * @see SocketManager
 * @author Dean Mitrovich
 * @version 1.0
 */
public class VVWNETCMDSENDER
{
    /** Millisecond time the last command was sent
     */
    public long dwCmdTimeMs = 0;
    /** Socket port to connect on
     */
    int m_dwPort = 1;
    /** Last error message
     */
    String m_szError = null;
    /**
     * InetAddress Class represents the IP address as an unsigned 32 bit number used by the IP
     */
    InetAddress m_inetAddr = null;
    /**
     * Used to connect to a foriegn host at a specified port
     */
    private Socket m_TCSocket = null;
    /**
     * The socket address
     */
    InetSocketAddress m_InetSocketAddress = null;
    /** Shared socket manager
     */
    SocketManager socketMgr = null;

    /** Input data stream used for communication through socket
     */
    DataInputStream m_DataInputStream = null;
    /** Output data stream used for communication through socket
     */
    DataOutputStream m_DataOutputStream = null;

    /** If true, then this sender is connected to a vvw and communicating
     */
    private boolean m_bValid = false;
    /** If true, then this sender's tcp/ip socket is connected to the remote machine
     */
    private boolean m_connected = false;
    /** If true, then the last command sent was sent are a response was recieved
     */
    boolean m_bLastCmdOK = false;
    /** If True, the sender is not ready to send another command. Normally
     * this is because it is already processing a command
     */
    private boolean m_NotReady = false;
    /** Total stream errors since connect
     */
    int dwStreamErrors = 0;
    /** Current channel the net sender is communicating with within the
     * remote vvw
     */
    private long m_Channel;
    /** Default time out.  Length in milliseconds to wait for a reponse
     * from the remote machine
     */
    private int m_timeOut = 10000; // 2000;
    /** Internal:  A unique number to attach to a command
     */
    private long m_dwSemiUnique = 1;
    /** Current date
     */
    Date m_Date = null;

    /**
     * Only for SocketLocket, do not use outside this function.  Shared
     * amoung ALL vvw net command senders
     */
    static private boolean m_fInUse_SocketLocket = false;

    /**
     * SocketLocket - Lock a socket for exclusive use
     *
     * @param fCheck boolean Do we want to acquire the lock, or just check if it is locked
     * @param fLock boolean Are we locking, or unlocking the socket
     * @return boolean true if locked
     */
    private synchronized boolean SocketLocket(boolean fCheck, boolean fLock)
    {
        // Just a check
        if(fCheck)
        {
            return m_fInUse_SocketLocket;
        }
        // It is an unlock
        if(fLock == false)
        {
            m_fInUse_SocketLocket = false;
            return m_fInUse_SocketLocket;
        }
        // Must be a lock
        if(m_fInUse_SocketLocket)
        {
            // Already locked, can not re lock
            return false;
        }
        m_fInUse_SocketLocket = true;
        return m_fInUse_SocketLocket;
    }

    /**
     * fInUse - Is the socket currently locked and in use
     *
     * @return boolean true if it is in use
     */
    boolean fInUse()
    {
        return SocketLocket(true, false);
    }

    /**
     * LockSocket - Lock the socket for exclusive use
     *
     * @return boolean true if it is in use
     */
    boolean LockSocket()
    {
        return SocketLocket(false, true);
    }

    /**
     * UnLockSocket - Unlock the socket, allow others access
     *
     * @return boolean true if it is in use
     */
    boolean UnLockSocket()
    {
        return SocketLocket(false, false);
    }

    /**
     * VVWNETCMDSENDER - Default, do not use
     */
    public VVWNETCMDSENDER()
    {
    }

    /**
     * VVWNETCMDSENDER - Creates sender and keeps reference to socket manager
     *
     * @param socketManager SocketManager Share socket resource for multiple
     * logical channels with only one physical pipe (XP socket limit problem)
     */
    public VVWNETCMDSENDER(SocketManager socketManager)
    {
        socketMgr = socketManager;
    }

    /**
     * Sets the current channel in use by the applet
     *
     * @param channel long
     */
    public void setChannel(long channel)
    {
        m_Channel = channel;
    }

    /**
     * getChannel
     *
     * @return long
     */
    public long getChannel()
    {
        return m_Channel;
    }

    /**
     * setTimeout for the low level socket
     *
     * @param nTimeout int
     */
    public void setTimeout(int nTimeout)
    {
        if(socketMgr == null)
        {
            m_timeOut = nTimeout;
            try
            {
                m_TCSocket.setSoTimeout(nTimeout);
            }
            catch(IOException e)
            {
                // Print any errors produced by above try statment
                System.out.println("\n\rIO Exception: " + e.getMessage());
            }
        }
        else
        {
            socketMgr.setTimeout(m_inetAddr, m_dwPort, nTimeout);
        }
    }

    /**
     * Attempts to establish a connection to the specified server at the specified port.
     *
     * @param szIpAddress String IP address or 'look up-able' name
     * @param dwPort int Normally 1234
     * @return boolean true is connection succeeds
     */
    public boolean TryConnect(String szIpAddress, int dwPort)
    {
        // If no server is specified or no port is specified then connection cannot be established
        if (szIpAddress == null || dwPort == 0xFFFFFFFF)
        {
            return false;
        }
        m_dwPort = dwPort;

        // Establish socket connection
        try
        {
            m_inetAddr = InetAddress.getByName(szIpAddress);

            if (m_inetAddr == null)
            {
                return false;
            }

            if (socketMgr == null)
            {
                // If the socket is currently in use the connection cannot be established
                if (m_TCSocket != null && m_connected == true)
                {
                    return true;
                }
                else
                {
                    m_TCSocket = null;
                }
                // Create the Socket to the specified server and port
                m_TCSocket = new Socket();
                m_InetSocketAddress = new InetSocketAddress(m_inetAddr, dwPort);
                m_TCSocket.connect(m_InetSocketAddress, m_timeOut);

                // Did we get a socket
                if (m_TCSocket == null || !m_TCSocket.isConnected())
                {
                    m_TCSocket = null;
                    Disconnect(false);
                    return false;
                }
                try
                {
                    // enable TCP_NODELAY
                    m_TCSocket.setTcpNoDelay(true);

                    // Set a read timeout
                    m_TCSocket.setSoTimeout(m_timeOut);
                }
                catch (IOException e)
                {
                    // Print any errors produced by above try statment
                    m_szError = e.getMessage();
                    System.out.println("\n\rIO Exception: " + m_szError);
                    m_TCSocket = null;
                    Disconnect(false);
                    return false;
                }

                // Establish Input and Output streams for communication (byte streams)
                try
                {
                    m_DataInputStream = new DataInputStream(m_TCSocket.getInputStream());
                    m_DataOutputStream = new DataOutputStream(m_TCSocket.getOutputStream());
                }
                catch (IOException e)
                {
                    // Print any errors produced by above try statment
                    m_szError = e.getMessage();
                    System.out.println("\n\rIO Exception: " + m_szError);
                    m_TCSocket = null;
                    Disconnect(false);
                    return false;
                }
            }
            else
            {
                if (socketMgr.Open(m_inetAddr, m_dwPort) == false)
                {
                    m_connected = false;
                    Disconnect(false);

                    return false;
                }

                // Establish Input and Output streams for communication (byte streams)
                try
                {
                   if((m_DataInputStream = new DataInputStream(socketMgr.GetInputStream(m_inetAddr, m_dwPort))) == null)
                   {
                       // GetInputStream Failed
                       dwStreamErrors ++;
                       Disconnect(false);
                       return false;
                   }
                   if((m_DataOutputStream = new DataOutputStream(socketMgr.GetOutputStream(m_inetAddr, m_dwPort))) == null)
                    {
                       // GetOutputStream Failed
                       dwStreamErrors ++;
                       Disconnect(false);
                       return false;
                   }
               }
               catch (Exception be)
               {
                   // Print any errors produced by above try statment
                    m_szError = be.getMessage();
                    System.out.println("\n\rIO Exception: " + m_szError);

                    // Can't create Input or Output Stream: Disconnect the socket
                    Disconnect(false);
                    return false;
                }
            }
        }
        catch (Exception be)
        {
            // Print any errors produced by above try statment
            m_szError = be.getMessage();
            System.out.println("Exception: " + m_szError);
            m_connected = false;
            Disconnect(false);
            return false;
        }
        // Establish date and time of connection and return succesfully
        m_Date = new Date();
        m_bValid = true;
        m_connected = true;
        return true;
    }

    /**
     * Disconnects the connection established by socket
     *
     * @param fCloseAllInst boolean Normally false, just close the instance
     * the caller opened.  If true, the actual socket will close.  This will
     * leave other parts of the program with bad handles to the old socket.
     */
    public void Disconnect(boolean fCloseAllInst)
    {
        if(IsConnected())
        {
            try
            {
                if (m_TCSocket != null)
                {
                    m_TCSocket.shutdownOutput();
                    m_TCSocket.shutdownInput();
                }
                if (socketMgr == null)
                {
                    if (m_DataOutputStream != null)
                    {
                        m_DataOutputStream.close();
                    }
                    if (m_DataInputStream != null)                   {
                        m_DataInputStream.close();
                    }

                    m_TCSocket.close();
                    if (m_TCSocket.isClosed())
                    {
                        System.out.println("\n\rSocket was successfully closed");
                        m_connected = false;
                    }
                    else
                    {
                        System.out.println("\n\rSocket was NOT successfully closed");
                    }
                }
                else
                {
                   if(socketMgr.Close(m_inetAddr, m_dwPort, fCloseAllInst))
                   {
                       m_connected = false;
                   }
                }
            }
            catch (Exception e)
            {
                m_szError = e.getMessage();
                System.out.println("\n\r Disconnect(): Exception: " + m_szError);
            }

            if (socketMgr == null)
            {
                // Set objects to null for garbage collection
                m_TCSocket = null;
            }
            m_DataInputStream = null;
            m_DataOutputStream = null;

            m_bValid = false;
            m_dwPort = -1;
            m_szError = null;
            m_inetAddr = null;
            m_bLastCmdOK = false;
            m_NotReady = true;
            m_Date = null;

        } // end: If
        else
        {
            System.out.println("\n\r Called Disconnect(), but it wasn't connected.");
        }
    } // end Disconnect()

    /**
     * BytesAvailable - number of bytes available in the input data stream
     *
     * @return int
     */
    private int BytesAvailable()
    {
        int nRtn = 0;
        try
        {
            nRtn = m_DataInputStream.available();
        }
        catch (SocketException e)
        {
            nRtn = 0;
        }
        catch (Exception ee)
        {
            nRtn = 0;
        }
        return nRtn;
    }

    /**
     * This sends a MediaCmd command to any channel in the VVW we are connected to/
     *
     * @param lChannel long The target channel 0..n
     * @param mCmd MEDIACMD The MedaiCmd to send
     * @return boolean true if the send is successful and returned
     */
    public synchronized boolean SendVvwCmd(long lChannel, MEDIACMD mCmd)
    {
        long dwStructureSize = 0;

        while(LockSocket() == false)
        {
            // Wait for socket lock
            try
            {
                Thread.sleep(1);
            } catch(Exception ee)
            {
                System.out.println("Failure: Line 390");
                return false;
            }
        }

        long dwCmdStartTimeMs = System.currentTimeMillis();
        long lTimeout = 50;
        if(!IsConnected())
        {
            mCmd.InitFailure();
            UnLockSocket();
            // System.out.println("Failure: Line 404 IsConnected() failed");
            return false;
        }
        m_bLastCmdOK = false; // set to false until command completed successfully

        mCmd.Channel(lChannel); // set the Channel

        // If the connection isn't established or the connection has been terminated
        // then the message is not sent
        if((m_bValid == false) || (m_NotReady == true))
        {
            mCmd.InitFailure();
            System.out.println("Message wasn't sent");
            m_bLastCmdOK = false;
            m_NotReady = false;
            UnLockSocket();
            if(m_bLastCmdOK)
            {
                // System.out.println("Success");
            }
            else
            {
                System.out.println("Failure: Line 426");
            }
            return m_bLastCmdOK;
        }
        else
        {
            // Don't accept any other commands until current command is done
            m_NotReady = true;
        }

        long dwParam = 0, dwStartTime = 0;
        int nBytesRtn = 0;

        // Checks to see if there is a command, if that command is the one currently
        // in use with the object mCmd of class MEDIACMD, and that the size of the
        // command is not to large for the object's max size

        if(mCmd == null || mCmd.CmdID() != MEDIACMD.MEDIACMD_CURRENT)
        {
            mCmd.InitFailure();
            m_NotReady = false;
            m_bLastCmdOK = false;
            UnLockSocket();
            if(m_bLastCmdOK)
            {
                // System.out.println("Success");
            }
            else
            {
                System.out.println("Failure: Line 456");
            }
            return m_bLastCmdOK;
        }

        // Sets the Struct size to the Base size of a media command
        dwStructureSize = mCmd.StructureSize();
        if(dwStructureSize == 0)
        {
            mCmd.StructureSize(MediaCmdIF.MEDIACMD_BASE_SIZE);
            dwStructureSize = MediaCmdIF.MEDIACMD_BASE_SIZE;
        }

        dwParam = mCmd.netTime(); // previous start time of mCmd
        dwStartTime = (m_Date.getTime() + m_dwSemiUnique * 59) / 37;
        m_dwSemiUnique += 1;
        mCmd.netTime(dwStartTime); // set the start time of the cmd
        dwStartTime = mCmd.netTime();

        // Try to write command to the server
        try
        {
            if(dwStructureSize <= MediaCmdIF.MEDIACMD_BASE_SIZE)
            {
                // Always write at least up to arbID
                mCmd.StructureSize(MediaCmdIF.MEDIACMD_BASE_SIZE);
                dwStructureSize = MediaCmdIF.MEDIACMD_BASE_SIZE;
                m_DataOutputStream.write(mCmd.byteDataBuffer, 0, MediaCmdIF.MEDIACMD_BASE_SIZE);
            }
            else
            {
                if(dwStructureSize < MEDIACMD.CMD_MAX_CLIP_ID_LEN && mCmd.byteDataBuffer[(int) dwStructureSize] != 0)
                {
                    mCmd.StructureSize((int) dwStructureSize + 1);
                    mCmd.byteDataBuffer[(int) dwStructureSize] = 0;
                }
                m_DataOutputStream.write(mCmd.byteDataBuffer, 0, (int) dwStructureSize);
            }
            // necessary to flush after use
            m_DataOutputStream.flush();
        } catch(SocketException e)
        {
            // Print any errors produced by above try statment
            m_szError = e.getMessage();
            System.out.println("*** Write Failed *** SocketException: " + m_szError);
            System.out.println(e.toString());

            mCmd.InitFailure();
            m_bLastCmdOK = false;
            m_NotReady = false;
            // Not sure  m_connected = false;

            if(m_szError.trim().compareTo("Software caused connection abort: socket write error") == 0)
            {
                Disconnect(true);
            }
            dwStreamErrors++;
            UnLockSocket();
            if(m_bLastCmdOK)
            {
                // System.out.println("Success");
            }
            else
            {
                System.out.println("Failure: Line 529");
            }
            return m_bLastCmdOK;
        } catch(Exception ee)
        {
            m_szError = ee.getMessage();
            System.out.println("\n\rIO Exception: " + m_szError);
        }

        // At this point the command is sent, we now need the return

        int ByteQty = 0;
        // DEBUG
        mCmd.netTime(0);
        mCmd.CmdID( -1);
        // Find the start of the returning structure.  The bytes will be as follows:
        //  ?? ?? ?? ??
        //  ?? ?? ?? ??
        //  dwStartTime
        //  mCmd.MEDIACMD_CURRENT
        //  dwStructureSize
        // Initially get 16+4 bytes, then shift one byte at a time until
        // the dwStartTime/mCmd.MEDIACMD_CURRENT match
        try
        {
            // Make sure we have at least 20 bytes of data available to cover all the
            // members we need to get the time/cmdid/size
            switch((int) mCmd.Cmd())
            {
                case MEDIACMD.ctTerminate:
                    Thread.sleep(5000);
                    lTimeout = 10000;
                    break;
                case MEDIACMD.ctTransfer:
                case MEDIACMD.ctDelete:
                case MEDIACMD.ctBlank:
                case MEDIACMD.ctTrim:
                    lTimeout = 10000;
                    break;
                case MEDIACMD.ctInsert:
                    lTimeout = 30000; //          30 sec
                    break;
                case MEDIACMD.ctSetValue:
                    switch((int) mCmd.CmdAlt())
                    {
                        case MEDIACMD.gsRecDrive:
                        case MEDIACMD.gsRecFileName:
                            lTimeout = 120000;
                            break;
                        case MEDIACMD.gsConvertFileFormat:
                        case MEDIACMD.gsConvertAudFileFormat:
                        case MEDIACMD.gsPiconFrame:
                        case MEDIACMD.gsJpegFrame:
                            lTimeout = 12000;
                            break;
                        case MEDIACMD.gsResetChannel:
                            lTimeout = 100000; // 100 seconds
                            break;
                        case MEDIACMD.gsSaveCurrent:
                        case MEDIACMD.gsLoadClipSpace:
                        case MEDIACMD.gsLoadTCSpace:
                        case MEDIACMD.gsLoadFilmSpace:
                            lTimeout = 30000; // 30 seconds
                            break;
                        default:
                            lTimeout = 500; // Max 1/4 second for return, should be lower
                            break;
                    }
                    break;
                default:
                    lTimeout = 500; // Max 1/4 second for return, should be lower
                    break;
            }

            //if(mCmd.Cmd() == mCmd.ctTransfer || mCmd.Cmd() == mCmd.ctTerminate)
            //{
                //boolean fDebug = true;
            //}

            while(lTimeout > 0 && BytesAvailable() < 20)
            {
                lTimeout--;
                Thread.sleep(1);
            }

            if(lTimeout <= 0 && BytesAvailable() < 20)
            {
                m_bLastCmdOK = false;
                m_NotReady = false;
                UnLockSocket();
                System.out.println("Failure: Line 591 TimedOut, mCmd.Cmd()= " + mCmd.Cmd());
                return false;
            }
        } catch(Exception ee)
        {
            //
            System.out.println("\n\rUnable to find first 20 bytes of return");
            m_bLastCmdOK = false;
            m_NotReady = false;
            UnLockSocket();
            if(m_bLastCmdOK)
            {
                // System.out.println("Success");
            }
            else
            {
                System.out.println("Failure: Line 608");
            }
            return m_bLastCmdOK;
        }

        // This will always get at least 20 bytes
        try
        {
            nBytesRtn = m_DataInputStream.read(mCmd.byteDataBuffer, 0, MediaCmdIF.MEDIACMD_BASE_SIZE);
        }
        catch(SocketException e)
        {
            System.out.println("\n\rLine 621, SocketException after m_DataInputStream.read(...)");
            System.out.println("\n\rSocketException->Initial mCmd Read (64 Bytes): " + e.toString());
            nBytesRtn = -1;
            // Not sure  m_connected = false;
        }
        catch(Exception ee)
        {
            System.out.println("\n\rLine 628, Exception after m_DataInputStream.read(...)");
            System.out.println("\n\rException->Initial mCmd Read (64 Bytes): " + ee.toString());
            nBytesRtn = -1;
        }

        // If we can't get at least 20 (mCmd.MEDIACMD_BASE_SIZE) bytes, fail and make mCmd invalid
        if(nBytesRtn < 20)
        {
            // Long time out.  Give up.
            System.out.println("\n\rRead socket failed with no characters (64 Bytes)");
            mCmd.InitFailure();
            m_bLastCmdOK = false;
            m_NotReady = false;
            dwStreamErrors++;

            UnLockSocket();
            if(m_bLastCmdOK)
            {
                // System.out.println("Success");
            }
            else
            {
                System.out.println("Failure: Line 653");
            }
            return m_bLastCmdOK;
        }

        // If these are not in the right place, then go through
        // the stream byte by byte to find the command return.
        dwStructureSize = 0;
        while(((mCmd.netTime() != dwStartTime) ||
            (mCmd.CmdID() != MEDIACMD.MEDIACMD_CURRENT) ||
            (mCmd.StructureSize() < MediaCmdIF.MEDIACMD_BASE_SIZE)) &&
            (BytesAvailable() > 0))
        {
            // We are NOT aligned - either
            //   Wrong id (dwStartTime)
            //   Bad cmd id (mediacmd_current)
            //   Struct size is too small
            if(nBytesRtn > 20)
            {
                // Move 1 byte back
                int i;
                for(i = 0; i < nBytesRtn; i++)
                {
                    mCmd.byteDataBuffer[i] = mCmd.byteDataBuffer[i + 1];
                }
                // Zero old byte
                mCmd.byteDataBuffer[i] = 0;
                nBytesRtn--;
            }
            // get up to mCmd.MEDIACMD_BASE_SIZE, but remove the 1 byte lost in the loop above
            try
            {
                // Read next set into buffer at offset
                nBytesRtn += m_DataInputStream.read(mCmd.byteDataBuffer, nBytesRtn,
                		MediaCmdIF.MEDIACMD_BASE_SIZE - nBytesRtn);
            }
            catch(Exception e)
            {
                System.out.println("Exception on mCmd read fill buffer");
            }
        }

        // At this point we have at least 20 bytes of the mediacmd return.  We
        // now need to collect the rest of the bytes
        if(nBytesRtn < 20)
        {
            // Ran out of return bytes, abort
            System.out.println("\n\rRead to align mCmd header ran out of bytes");
            System.out.println("\n\rRan out of return bytes, nBytesRtn= " + nBytesRtn +
                "  BytesAvailable= " + BytesAvailable());
            mCmd.InitFailure();
            m_bLastCmdOK = false;
            m_NotReady = false;
            UnLockSocket();
            if(m_bLastCmdOK)
            {
                // System.out.println("Success");
            }
            else
            {
                System.out.println("Failure: Line 712");
            }
            return m_bLastCmdOK;
        }

        // We have a properly aligned command
        ByteQty = nBytesRtn; // may be 20 < nBytesRtn <= mCmd.MEDIACMD_BASE_SIZE;
        nBytesRtn = 0;
        dwStructureSize = mCmd.StructureSize();

        // This should not be possible now.  The most bytes we can have at this
        // point would be mCmd.MEDIACMD_BASE_SIZE
        if(dwStructureSize > mCmd.byteDataBuffer.length)
        {
            System.out.println("\n\rThis Should Be Impossible - Error: Structure Size: " + dwStructureSize +
                " Buffer Size: " + mCmd.byteDataBuffer.length + ", mCmd.Cmd()= " + mCmd.Cmd());

            mCmd.StructureSize(mCmd.byteDataBuffer.length);
            dwStructureSize = mCmd.byteDataBuffer.length;
        }

        lTimeout = 100;

        // Get the remaining bytes in the command
        while(ByteQty < dwStructureSize && lTimeout > 0)
        {
            lTimeout--;
            // Get the remaining bytes -
            // Start of buffer = ByteQty
            // Remaining bytes = dwStructureSize - ByteQty (total - rx'd)
            try
            {
                nBytesRtn = m_DataInputStream.read(mCmd.byteDataBuffer, ByteQty, (int) dwStructureSize - ByteQty);
            }
            catch(SocketException e)
            {
                System.out.println("\n\rSocketException->ReadLoop (" + (long) (dwStructureSize - ByteQty) +
                    " Bytes): " + e.toString());
                nBytesRtn = -1;

                System.out.println("\n\rLine 755, mCmd.StructureSize() = " + dwStructureSize + ", ByteQty = " + ByteQty +
                    ", (mCmd.StructureSize() - ByteQty) = " + (dwStructureSize - ByteQty));
            }
            catch(Exception ee)
            {
                System.out.println("\n\rException->ReadLoop (" + (long) (dwStructureSize - ByteQty) + " Bytes): " +
                    ee.toString());
                nBytesRtn = -1;

                System.out.println("\n\rLine 764, mCmd.StructureSize() = " + dwStructureSize + ", ByteQty = " + ByteQty +
                    ", (mCmd.StructureSize() - ByteQty) = " + (dwStructureSize - ByteQty));
            }

            if(nBytesRtn != -1)
            {
                ByteQty = ByteQty + nBytesRtn;
                if(BytesAvailable() <= 0 && ByteQty < dwStructureSize)
                {
                    try
                    {
                        Thread.sleep(1); // *** Sleep long enough to ensure Update has completed its cycle after flag was set
                    }
                    catch(InterruptedException ex)
                    {
                        // will only happen if you cause the interruption
                    }
                }
            }
            else
            {
                // We have a -1 (exception from read)
                nBytesRtn = 0; // Try again
                try
                {
                    Thread.sleep(5);
                } catch(InterruptedException ex)
                {
                    // will only happen if you cause the interruption
                }
            }
            if(nBytesRtn <= 0)
            {
                lTimeout = 0;
            }
        }

        if(ByteQty != dwStructureSize)
        {
            System.out.println("Incorrect structure size >< returned bytes");
            m_bLastCmdOK = true;
            m_NotReady = false;
            UnLockSocket();
            if(m_bLastCmdOK)
            {
            }
            else
            {
                System.out.println("Failure: bLastCmd NOT Ok 882");
            }
            return true;
        }

        mCmd.netTime(dwParam);
        m_bLastCmdOK = true;
        m_NotReady = false;

        dwCmdTimeMs = System.currentTimeMillis() - dwCmdStartTimeMs;
        UnLockSocket();

        if(m_bLastCmdOK)
        {
        }
        else
        {
            System.out.println("Failure: bLastCmd NOT Ok 900");
        }
        return m_bLastCmdOK;
    }

    /**
     * Sends a command to the current channel
     *
     * @param mCmd MEDIACMD
     * @return boolean
     */
    public synchronized boolean SendVvwCmd(MEDIACMD mCmd)
    {
        return this.SendVvwCmd(m_Channel, mCmd);
    }

    /**
     * IsConnected - Are we currently connected through the socket
     *
     * @return boolean true is we are connected
     */
    public boolean IsConnected()
    {
        if (!m_connected)
        {
            return false;
        }
        else
        {
            if (socketMgr == null)
            {
                if (m_TCSocket == null)
                {
                    m_connected = false;
                    return false;
                }
                // Flag says it was connected but check
                // in case it was disconnected from the server side
                if (m_TCSocket.isConnected())
                {
                    return true;
                }
                else
                {
                    m_connected = false;
                    return false;
                }
            }
            else
            {
                if (socketMgr.IsSocketConnected(m_inetAddr, m_dwPort))
                {
                    return true;
                }
                else
                {
                    m_connected = false;
                    return false;
                }
            }
        }
    }

    /**
     * IsValid - Is our connection currently valid
     *
     * @return boolean true if it is valid
     */
    public boolean IsValid()
    {
        return this.m_bValid;
    }

    /**
     * IsReady - Is our connection ready to receive commands
     *
     * @return boolean true if it is
     */
    public boolean IsReady()
    {
        return (!m_NotReady);
    }

    /**
     * StreamErrors - Total stream errors since we first instantiated
     *
     * @return int Number of errors
     */
    public int StreamErrors()
    {
        return dwStreamErrors;
    }

    /**
     * IsFirstConnect - Is this the first time we have connected to this
     * machine, or are we going to reuse a currently connected socket.
     *
     * @return boolean true is we are not connected
     */
    public boolean IsFirstConnect()
    {
        if(socketMgr == null)
        {
            return true;
        }
        else
        {
            try
            {
                if(socketMgr.Find(m_inetAddr, m_dwPort).fFirstConnect)
                {
                    // First instance of this
                    return true;
                }
            }
            catch(Exception e)
            {
                return false;
            }
        }

        return false;
    }
}
