package packetNetworking;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Server class. May connecto to several clients. Extends the communicator class
 * @author Stian
 */
public class Server extends Communicator {
    
    /**
     * lists of ports, sockets, sender objects and client sockets
     */
    protected int[] ports;
	ServerSocket[] serverSockets;
	ObjectSender[] senders;
	Socket[] clientSockets;
    /**
     * true only if all clients is connected
     */
    protected boolean isAllSocketsOpen = false;
    /**
     * counts the number of ping errors. Resets when a correct ping is received. Status is unconnected if three pings fail
     */
    protected int pingErrorCounter = 0;
    /**
     * list of open clients. every bool is true only if corresponding client is open
     */
    protected boolean[] clientSocketsOpen;
    /**
     * true if ping is sent corresponding client
     */
    protected boolean[] hasSentPing;

    //Threads and listeners
    TimeoutThread[] timeoutThreads;
	ClientListenerThread[] clientListeners;
	MessageListenerThread[] messageListeners;

    //Listens for clients until contact is established
    ClientStatusListener clientStatusListener;
	
    /**
     * Sets up a server waiting to connect on given ports
     * @param ports
     */
    public Server(int[] ports) {
		
		super();
		
		this.ports = ports;
		
		serverSockets = new ServerSocket[ports.length];
		outs = new PrintStream[ports.length];
		ins = new BufferedInputStream[ports.length];
		senders = new ObjectSender[ports.length];
        clientSocketsOpen = new boolean[ports.length];
        hasSentPing = new boolean[ports.length];
        timeoutThreads = new TimeoutThread[ports.length];
		clientSockets = new Socket[ports.length];
		messageListeners = new MessageListenerThread[ports.length];
		clientListeners = new ClientListenerThread[ports.length];
		
		for(int i=0;i<ports.length;i++) {
			
			try {
				serverSockets[i] = new ServerSocket(ports[i]);
			} catch (IOException e) {
				e.printStackTrace();
			}
			
			System.out.println("ServerSocket "+(i+1)+" open");
			
			messageListeners[i] = new MessageListenerThread(this, i);
			
			clientListeners[i] = new ClientListenerThread(this, i);
			clientListeners[i].start();

            clientSocketsOpen[i] = false;
		}

        clientStatusListener = new ClientStatusListener();
        clientStatusListener.start();
		
	}

    /**
     * Process ping packets only. To use other packets, classes can overwrite this method
     * @param pck
     */
    public void ProcessPacket(Packet pck) {
        if(pck == null) return;
        if(pck instanceof StringPacket) {
            StringPacket sp = (StringPacket)pck;
            if(sp.text.substring(0,4).matches("ping")) {
                //Is ping. Process and update status of server
                int index = Integer.parseInt(sp.text.substring(4));
                if(hasSentPing[index]) {
                    hasSentPing[index] = false;
                    timeoutThreads[index].stop = true;
                    UpdateStatus(index);
                }
            }
        }
	}

    /*
     * Sends a packet to all clients
     *
     * */
    @Override
	public void Talk(Packet pck) {
		for(int i=0;i<ports.length;i++) {
            if(clientSockets[i]!=null) {
                if(senders[i]!=null) {
                    senders[i].SendObject(pck);
                }
            } else {
                System.out.println("client socket "+i+" is not connected");
            }
			
		}
		
	}

    /**
     * Sends a packet to a given client
     * @param command
     * @param index
     */
    public void Talk(Packet pck, int index) {
        if(senders[index]!=null) senders[index].SendObject(pck);
    }
	
    /**
     * Start listening from clients
     */
    public void Listen() {
		for(int i=0;i<ports.length;i++) {
			messageListeners[i].start();
		}
	}

    /**
     * Update connection status of the server
     * @param index
     */
    public void UpdateStatus(int index) {
        if(index>=0) {
            boolean s = true;
            clientSocketsOpen[index] = true;
            for(int i=0;i<ports.length;i++) {
                if(!clientSocketsOpen[i]) {
                    s = false;
                }
            }
            if(s) {
                isAllSocketsOpen = s;
            } else {
                pingErrorCounter++;
            }
        } else {
            pingErrorCounter++;
            if(pingErrorCounter<3) {
                return;
            }
            isAllSocketsOpen = false;
            clientSocketsOpen[-index-1] = false;
            pingErrorCounter = 0;
        }
        
        
    }

    //Class that listens for clients to connect
    private class ClientListenerThread extends Thread {

        //Access to the server itself to be able to connect
        Server server;
        int index;

        public ClientListenerThread(Server server, int index) {

            this.server = server;
            this.index = index;

        }

        @Override
        public void run() {

            while(true) {
                //Wait until timeout or client connects
                try {
                    server.clientSockets[index] = server.serverSockets[index].accept();
                } catch (IOException e) {
                    e.printStackTrace();
                }

                System.out.println("clientsocket up");

                //Open communication with and establish a sender object
                try {
                    server.outs[index] = new PrintStream(server.clientSockets[index].getOutputStream(), true);
                    server.ins[index] = new BufferedInputStream(server.clientSockets[index].getInputStream());
                    server.senders[index] = new ObjectSender(server.outs[index]);
                } catch (IOException e) {
                    e.printStackTrace();
                }

                //Setup a message listener
                messageListeners[index].stop = true;
                messageListeners[index] = new MessageListenerThread(server, index);
                messageListeners[index].start();
                
                System.out.println("Server up");
                try {
                    Thread.sleep(30);
                } catch (InterruptedException ex) {
                    Logger.getLogger(Server.class.getName()).log(Level.SEVERE, null, ex);
                }
            }
        }
    }

    //Thread that maintains connection status from clients. A ping is sent every second and waits for response with a timeout thread
    private class ClientStatusListener extends Thread {

        @Override
        public void run() {
            while(true) {
                for(int i=0;i<ports.length;i++) {
                    Talk(new StringPacket("ping"+i));
                    hasSentPing[i] = true;

                    //Establishes a timeout thread
                    timeoutThreads[i] = new TimeoutThread(i);
                    timeoutThreads[i].start();
                }
                try {
                    Thread.sleep(1000);
                    //TODO: Change interval
                } catch (InterruptedException ex) {
                    Logger.getLogger(Server.class.getName()).log(Level.SEVERE, null, ex);
                }
            }
        }

    }


    //A timeout thread. Waits for 1500ms before setting negative status on the corresponding client
    private class TimeoutThread extends Thread {

        int index;
        boolean stop;

        public TimeoutThread(int index) {
            this.index = index;
            stop = false;
        }

        @Override
        public void run() {
            try {
                Thread.sleep(1500);
                if(!stop) UpdateStatus(-index - 1);
            } catch (InterruptedException ex) {
                Logger.getLogger(Server.class.getName()).log(Level.SEVERE, null, ex);
            }
        }

    }

}
